import re
from datetime import datetime, timezone
from importlib.metadata import version
from pathlib import Path
from typing import Iterable, Optional

from jproperties import Properties
from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator, model_validator
from typing_extensions import Self
from upath import UPath

from hats.catalog.catalog_type import CatalogType
from hats.io import file_io

## catalog_name, catalog_type, and total_rows are required for ALL types
CATALOG_TYPE_REQUIRED_FIELDS = {
    CatalogType.OBJECT: ["ra_column", "dec_column"],
    CatalogType.SOURCE: ["ra_column", "dec_column"],
    CatalogType.ASSOCIATION: [
        "primary_catalog",
        "primary_column",
        "join_catalog",
        "join_column",
        "contains_leaf_files",
    ],
    CatalogType.INDEX: ["primary_catalog", "indexing_column"],
    CatalogType.MARGIN: ["primary_catalog", "margin_threshold"],
    CatalogType.MAP: [],
}


class TableProperties(BaseModel):
    """Container class for catalog metadata"""

    catalog_name: str = Field(alias="obs_collection")
    catalog_type: CatalogType = Field(alias="dataproduct_type")
    total_rows: int = Field(alias="hats_nrows")

    ra_column: Optional[str] = Field(default=None, alias="hats_col_ra")
    dec_column: Optional[str] = Field(default=None, alias="hats_col_dec")
    default_columns: Optional[list[str]] = Field(default=None, alias="hats_cols_default")
    """Which columns should be read from parquet files, when user doesn't otherwise specify."""

    primary_catalog: Optional[str] = Field(default=None, alias="hats_primary_table_url")
    """Reference to object catalog. Relevant for nested, margin, association, and index."""

    margin_threshold: Optional[float] = Field(default=None, alias="hats_margin_threshold")
    """Threshold of the pixel boundary, expressed in arcseconds."""

    primary_column: Optional[str] = Field(default=None, alias="hats_col_assn_primary")
    """Column name in the primary (left) side of join."""

    primary_column_association: Optional[str] = Field(default=None, alias="hats_col_assn_primary_assn")
    """Column name in the association table that matches the primary (left) side of join."""

    join_catalog: Optional[str] = Field(default=None, alias="hats_assn_join_table_url")
    """Catalog name for the joining (right) side of association."""

    join_column: Optional[str] = Field(default=None, alias="hats_col_assn_join")
    """Column name in the joining (right) side of join."""

    join_column_association: Optional[str] = Field(default=None, alias="hats_col_assn_join_assn")
    """Column name in the association table that matches the joining (right) side of join."""

    contains_leaf_files: Optional[bool] = Field(default=None, alias="hats_assn_leaf_files")
    """Whether or not the association catalog contains leaf parquet files."""

    indexing_column: Optional[str] = Field(default=None, alias="hats_index_column")
    """Column that we provide an index over."""

    extra_columns: Optional[list[str]] = Field(default=None, alias="hats_index_extra_column")
    """Any additional payload columns included in index."""

    npix_suffix: str = Field(default=".parquet", alias="hats_npix_suffix")
    """Suffix of the Npix partitions.
    In the standard HATS directory structure, this is '.parquet' because there is a single file
    in each Npix partition and it is named like 'Npix=313.parquet'.
    Other valid directory structures include those with the same single file per partition but
    which use a different suffix (e.g., `npix_suffix` = '.parq' or '.snappy.parquet'),
    and also those in which the Npix partitions are actually directories containing 1+ files
    underneath (and then `npix_suffix` = '/').
    """

    skymap_order: Optional[int] = Field(default=None, alias="hats_skymap_order")
    """Nested Order of the healpix skymap stored in the default skymap.fits."""

    skymap_alt_orders: Optional[list[int]] = Field(default=None, alias="hats_skymap_alt_orders")
    """Nested Order (K) of the healpix skymaps stored in altnernative skymap.K.fits."""

    ## Allow any extra keyword args to be stored on the properties object.
    model_config = ConfigDict(extra="allow", populate_by_name=True, use_enum_values=True)

    @field_validator("default_columns", "extra_columns", mode="before")
    @classmethod
    def space_delimited_list(cls, str_value: str) -> list[str]:
        """Convert a space-delimited list string into a python list of strings."""
        if isinstance(str_value, str):
            # Split on a few kinds of delimiters (just to be safe), and remove duplicates
            return list(filter(None, re.split(";| |,|\n", str_value)))
        ## Convert empty strings and empty lists to None
        return str_value if str_value else None

    @field_validator("skymap_alt_orders", mode="before")
    @classmethod
    def space_delimited_int_list(cls, str_value: str | list[int]) -> list[int]:
        """Convert a space-delimited list string into a python list of integers.

        Args:
            str_value(str | list[int]): string representation of a list of integers, delimited by
                space, comma, or semi-colon, or a list of integers.

        Returns:
            sorted list of unique integers, if all inputs are integers, None if the input is empty
        Raises:
            ValueError: if any non-digit characters are encountered, or
        """
        if not str_value:
            return None
        if isinstance(str_value, int):
            return [str_value]
        if isinstance(str_value, str):
            # Split on a few kinds of delimiters (just to be safe)
            int_list = [int(token) for token in list(filter(None, re.split(";| |,|\n", str_value)))]
        elif isinstance(str_value, list) and all(isinstance(elem, int) for elem in str_value):
            int_list = str_value
        else:
            raise ValueError(f"Unsupported type of skymap_alt_orders {type(str_value)}")
        if len(int_list) == 0:
            return None
        int_list = list(set(int_list))
        int_list.sort()
        return int_list

    @field_serializer("default_columns", "extra_columns", "skymap_alt_orders")
    def serialize_as_space_delimited_list(self, str_list: Iterable) -> str:
        """Convert a python list of strings into a space-delimited string."""
        if str_list is None or len(str_list) == 0:
            return None
        return " ".join([str(element) for element in str_list])

    @model_validator(mode="after")
    def check_required(self) -> Self:
        """Check that type-specific fields are appropriate, and required fields are set."""
        explicit_keys = set(
            self.model_dump(by_alias=False, exclude_none=True).keys() - self.__pydantic_extra__.keys()
        )

        required_keys = set(
            CATALOG_TYPE_REQUIRED_FIELDS[self.catalog_type] + ["catalog_name", "catalog_type", "total_rows"]
        )
        missing_required = required_keys - explicit_keys
        if len(missing_required) > 0:
            raise ValueError(
                f"Missing required property for table type {self.catalog_type}: {missing_required}"
            )

        return self

    def copy_and_update(self, **kwargs):
        """Create a validated copy of these table properties, updating the fields provided in kwargs."""
        new_properties = self.model_copy(update=kwargs)
        TableProperties.model_validate(new_properties)
        return new_properties

    def explicit_dict(self, by_alias=False, exclude_none=True):
        """Create a dict, based on fields that have been explicitly set, and are not "extra" keys."""
        explicit = self.model_dump(by_alias=by_alias, exclude_none=exclude_none)
        extra_keys = self.__pydantic_extra__.keys()
        return {key: val for key, val in explicit.items() if key not in extra_keys}

    def extra_dict(self, by_alias=False, exclude_none=True):
        """Create a dict, based on fields that are "extra" keys."""
        explicit = self.model_dump(by_alias=by_alias, exclude_none=exclude_none)
        extra_keys = self.__pydantic_extra__.keys()
        return {key: val for key, val in explicit.items() if key in extra_keys}

    def __str__(self):
        """Friendly string representation based on named fields."""
        parameters = self.explicit_dict()
        formatted_string = ""
        for name, value in parameters.items():
            formatted_string += f"  {name} {value}\n"
        return formatted_string

    @classmethod
    def read_from_dir(cls, catalog_dir: str | Path | UPath) -> Self:
        """Read field values from a java-style properties file."""
        catalog_path = file_io.get_upath(catalog_dir)
        file_path = catalog_path / "hats.properties"
        if not file_io.does_file_or_directory_exist(file_path):
            file_path = catalog_path / "properties"
            if not file_io.does_file_or_directory_exist(file_path):
                raise FileNotFoundError(f"No properties file found where expected: {str(file_path)}")
        p = Properties()
        with file_path.open("rb") as f:
            p.load(f, "utf-8")
        return cls(**p.properties)

    def to_properties_file(self, catalog_dir: str | Path | UPath) -> Self:
        """Write fields to a java-style properties file."""
        # pylint: disable=protected-access
        parameters = self.model_dump(by_alias=True, exclude_none=True)
        properties = Properties()
        properties.properties = parameters
        properties._key_order = parameters.keys()

        catalog_path = file_io.get_upath(catalog_dir)
        file_path = catalog_path / "hats.properties"
        with file_path.open("wb") as _file:
            properties.store(_file, encoding="utf-8", initial_comments="HATS catalog", timestamp=False)
        file_path = catalog_path / "properties"
        with file_path.open("wb") as _file:
            properties.store(_file, encoding="utf-8", initial_comments="HATS catalog", timestamp=False)

    @staticmethod
    def new_provenance_dict(
        path: str | Path | UPath | None = None, builder: str | None = None, **kwargs
    ) -> dict:
        """Constructs the provenance properties for a HATS catalog.

        Args:
            path (UPath): The path to the catalog directory.
            builder (str | None): The name and version of the tool that created the catalog.
            **kwargs: Additional properties to include/override in the dictionary.

        Returns:
            A dictionary with properties for the HATS catalog.
        """

        def _estimate_dir_size(target_dir):
            total_size = 0
            for item in target_dir.iterdir():
                if item.is_dir():
                    total_size += _estimate_dir_size(item)
                else:
                    total_size += item.stat().st_size
            return total_size

        path = file_io.get_upath(path)

        builder_str = ""
        if builder is not None:
            builder_str = f"{builder}, "
        builder_str += f"hats v{version('hats')}"

        properties = {}
        now = datetime.now(tz=timezone.utc)
        properties["hats_builder"] = builder_str
        properties["hats_creation_date"] = now.strftime("%Y-%m-%dT%H:%M%Z")
        properties["hats_estsize"] = int(_estimate_dir_size(path) / 1024) if path else 0
        properties["hats_release_date"] = "2024-09-18"
        properties["hats_version"] = "v0.1"
        return properties | kwargs



from __future__ import annotations

from upath import UPath

from hats.catalog import Catalog
from hats.catalog.dataset.collection_properties import CollectionProperties
from hats.pixel_math import HealpixPixel


class CatalogCollection:
    """A collection of HATS Catalog with data stored in a HEALPix Hive partitioned structure

    Catalogs of this type are described by a `collection.properties` file which specifies
    the underlying main catalog, margin catalog and index catalog paths. These catalogs are
    stored at the root of the collection, each in its separate directory::

        catalog_collection/
        ├── main_catalog/
        ├── margin_catalog/
        ├── index_catalog/
        ├── collection.properties

    Margin and index catalogs are optional but there could also be multiple of them. The
    catalogs used by default are specified in the `collection.properties` file in the
    `default_margin` and `default_index` keywords.
    """

    def __init__(
        self,
        collection_path: UPath,
        collection_properties: CollectionProperties,
        main_catalog: Catalog,
    ):
        self.collection_path = collection_path
        self.collection_properties = collection_properties

        if not isinstance(main_catalog, Catalog):
            raise TypeError(f"HATS at {main_catalog.catalog_path} is not of type `Catalog`")
        self.main_catalog = main_catalog

    @property
    def main_catalog_dir(self) -> UPath:
        """Path to the main catalog directory"""
        return self.collection_path / self.collection_properties.hats_primary_table_url

    @property
    def all_margins(self) -> list[str] | None:
        """The list of margin catalog names in the collection"""
        return self.collection_properties.all_margins

    @property
    def default_margin(self) -> str | None:
        """The name of the default margin"""
        return self.collection_properties.default_margin

    @property
    def default_margin_catalog_dir(self) -> UPath | None:
        """Path to the default margin catalog directory"""
        if self.default_margin is None:
            return None
        return self.collection_path / self.default_margin

    @property
    def all_indexes(self) -> dict[str, str] | None:
        """The mapping of indexes in the collection"""
        return self.collection_properties.all_indexes

    @property
    def default_index_field(self) -> str | None:
        """The name of the default index field"""
        return self.collection_properties.default_index

    @property
    def default_index_catalog_dir(self) -> UPath | None:
        """Path to the default index catalog directory"""
        if self.default_index_field is None:
            return None
        default_index_dir = self.all_indexes[self.default_index_field]
        return self.collection_path / default_index_dir

    def get_index_dir_for_field(self, field_name: str | None = None) -> UPath | None:
        """Path to the field's index catalog directory"""
        if field_name is None:
            return self.default_index_catalog_dir
        if self.all_indexes is None or field_name not in self.all_indexes:
            raise ValueError(f"Index for field `{field_name}` is not specified in all_indexes")
        index_dir = self.all_indexes[field_name]
        return self.collection_path / index_dir

    def get_healpix_pixels(self) -> list[HealpixPixel]:
        """The list of HEALPix pixels of the main catalog"""
        return self.main_catalog.get_healpix_pixels()



import pytest

from hats.catalog.dataset.collection_properties import CollectionProperties
from hats.io.file_io import write_string_to_file
from hats.io.file_io.file_io import load_text_file


def test_read_collection_from_file(small_sky_collection_dir):
    properties_from_file = CollectionProperties.read_from_dir(small_sky_collection_dir).model_dump(
        by_alias=False, exclude_none=True
    )

    expected_attributes = {
        "name": "small_sky_o1_collection",
        "hats_primary_table_url": "small_sky_order1",
        "all_margins": "small_sky_order1_margin",
        "default_margin": "small_sky_order1_margin",
        "all_indexes": "id small_sky_order1_id_index",
        "default_index": "id",
        "obs_regime": "Optical",
    }

    for key, val in expected_attributes.items():
        assert properties_from_file[key] == val


def test_collection_properties_string():
    expected_properties = CollectionProperties(
        name="small_sky_01",
        hats_primary_table_url="small_sky_order1",
        obs_regime="Optical",
    )

    ## str representation should not include additional properties.
    assert (
        str(expected_properties)
        == """  name small_sky_01
  hats_primary_table_url small_sky_order1
"""
    )


def test_read_collection_list_parse(tmp_path):
    test_file_path = tmp_path / "collection.properties"

    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  all_margins=small_sky_order1_margin small_sky_order1_margin_10arcs
  all_indexes=
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    parsed_properties = CollectionProperties.read_from_dir(tmp_path)
    assert parsed_properties.all_margins == ["small_sky_order1_margin", "small_sky_order1_margin_10arcs"]
    assert parsed_properties.all_indexes is None

    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  all_margins=
  all_indexes=id small_sky_order1_id_index
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    parsed_properties = CollectionProperties.read_from_dir(tmp_path)
    assert parsed_properties.all_margins is None
    assert parsed_properties.all_indexes == {"id": "small_sky_order1_id_index"}

    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  all_margins=
  all_indexes=
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    parsed_properties = CollectionProperties.read_from_dir(tmp_path)

    parsed_properties.to_properties_file(tmp_path)

    contents = load_text_file(test_file_path)
    assert contents == [
        "#HATS Collection\n",
        "obs_collection=foo\n",
        "hats_primary_table_url=small_sky_order1\n",
    ]


def test_write_properties_colon_not_escaped(tmp_path):
    test_timestamp_value = "2025-06-30T17:10UTC"

    collection_props = CollectionProperties(
        name="test_collection_with_colon",
        hats_primary_table_url="test_url_with_colon",
        hats_creation_date=test_timestamp_value,
    )

    collection_props.to_properties_file(tmp_path)

    output_file_path = tmp_path / "collection.properties"

    file_contents = load_text_file(output_file_path)

    expected_line_unescaped = f"hats_creation_date={test_timestamp_value}\n"
    assert expected_line_unescaped in file_contents

    escaped_timestamp_value = test_timestamp_value.replace(":", r"\:")
    unexpected_line_escaped = f"hats_creation_date={escaped_timestamp_value}\n"
    assert unexpected_line_escaped not in file_contents


def test_read_collection_read_errors(tmp_path):
    test_file_path = tmp_path / "collection.properties"

    with pytest.raises(FileNotFoundError):
        CollectionProperties.read_from_dir(tmp_path)

    ## Very basic content - should work.
    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    _ = CollectionProperties.read_from_dir(tmp_path)

    ## Unsupported extra property.
    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  hats_col_ra=ra
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    with pytest.raises(ValueError, match="hats_col_ra"):
        CollectionProperties.read_from_dir(tmp_path)

    ## Unmatched index map.
    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  all_indexes=ra
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    with pytest.raises(ValueError, match="all_indexes"):
        CollectionProperties.read_from_dir(tmp_path)


def test_read_collection_from_file_round_trip(small_sky_collection_dir, tmp_path):
    table_properties = CollectionProperties.read_from_dir(small_sky_collection_dir)
    table_properties.to_properties_file(tmp_path)
    round_trip_properties = CollectionProperties.read_from_dir(tmp_path)

    assert table_properties == round_trip_properties

    kwarg_properties = CollectionProperties(
        **round_trip_properties.model_dump(by_alias=False, exclude_none=True)
    )
    assert table_properties == kwarg_properties


def test_read_collection_default_margin_not_found(small_sky_collection_dir, tmp_path):
    table_properties = CollectionProperties.read_from_dir(small_sky_collection_dir)
    assert table_properties.all_margins == ["small_sky_order1_margin"]
    table_properties.default_margin = "small_sky_order1_margin_1deg"
    table_properties.to_properties_file(tmp_path)
    with pytest.raises(ValueError, match="not found in all_margins"):
        CollectionProperties.read_from_dir(tmp_path)


def test_read_collection_all_margins_not_specified(small_sky_collection_dir, tmp_path):
    table_properties = CollectionProperties.read_from_dir(small_sky_collection_dir)
    assert table_properties.default_margin == "small_sky_order1_margin"
    table_properties.all_margins = []
    table_properties.to_properties_file(tmp_path)
    with pytest.raises(ValueError, match="all_margins needs to be set"):
        CollectionProperties.read_from_dir(tmp_path)


def test_read_collection_default_index_not_found(small_sky_collection_dir, tmp_path):
    table_properties = CollectionProperties.read_from_dir(small_sky_collection_dir)
    assert table_properties.all_indexes == {"id": "small_sky_order1_id_index"}
    table_properties.default_index = "obj_name"
    table_properties.to_properties_file(tmp_path)
    with pytest.raises(ValueError, match="not found in all_indexes"):
        CollectionProperties.read_from_dir(tmp_path)


def test_read_collection_all_indexes_not_specified(small_sky_collection_dir, tmp_path):
    table_properties = CollectionProperties.read_from_dir(small_sky_collection_dir)
    assert table_properties.default_index == "id"
    table_properties.all_indexes = []
    table_properties.to_properties_file(tmp_path)
    with pytest.raises(ValueError, match="all_indexes needs to be set"):
        CollectionProperties.read_from_dir(tmp_path)


def test_collection_parsing():
    ## Confirm we can pass in already dict- or list-like objects, and get the expected values.
    ## Space-only delimited
    results = CollectionProperties.space_delimited_list(
        "small_sky_order1_margin small_sky_order1_50arcs_margin"
    )
    assert results == ["small_sky_order1_margin", "small_sky_order1_50arcs_margin"]

    ## Comma-delimited
    results = CollectionProperties.space_delimited_list(
        "small_sky_order1_margin, small_sky_order1_50arcs_margin"
    )
    assert results == ["small_sky_order1_margin", "small_sky_order1_50arcs_margin"]

    ## Already a list of strings
    results = CollectionProperties.space_delimited_list(
        ["small_sky_order1_margin", "small_sky_order1_50arcs_margin"]
    )
    assert results == ["small_sky_order1_margin", "small_sky_order1_50arcs_margin"]

    ## None or empty
    results = CollectionProperties.space_delimited_list(None)
    assert results is None

    results = CollectionProperties.space_delimited_list("")
    assert results is None

    results = CollectionProperties.index_tuples("id small_sky_order1_id_index")
    assert results == {"id": "small_sky_order1_id_index"}

    ## Comma-delimited
    results = CollectionProperties.index_tuples("id small_sky_order1_id_index, ra small_sky_ra_index")
    assert results == {"id": "small_sky_order1_id_index", "ra": "small_sky_ra_index"}

    ## Already a dict of strings
    results = CollectionProperties.index_tuples(
        {"id": "small_sky_order1_id_index", "ra": "small_sky_ra_index"}
    )
    assert results == {"id": "small_sky_order1_id_index", "ra": "small_sky_ra_index"}

    ## None or empty
    results = CollectionProperties.index_tuples(None)
    assert results is None

    results = CollectionProperties.index_tuples("")
    assert results is None

    simple_properties = CollectionProperties(
        name="small_sky_01",
        hats_primary_table_url="small_sky_order1",
    )

    simple_properties_with_none = CollectionProperties(
        name="small_sky_01",
        hats_primary_table_url="small_sky_order1",
        all_margins=None,
        default_margin=None,
        all_indexes=None,
        default_index=None,
    )

    assert simple_properties == simple_properties_with_none



"""Utility functions for handling parquet metadata files"""

from __future__ import annotations

import random
from pathlib import Path

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as pds
import pyarrow.parquet as pq
from upath import UPath

from hats.io import file_io, paths
from hats.io.file_io.file_pointer import get_upath
from hats.pixel_math.healpix_pixel import HealpixPixel
from hats.pixel_math.healpix_pixel_function import get_pixel_argsort
from hats.pixel_math.spatial_index import SPATIAL_INDEX_COLUMN


# pylint: disable=too-many-locals
def write_parquet_metadata(
    catalog_path: str | Path | UPath,
    order_by_healpix=True,
    output_path: str | Path | UPath | None = None,
    create_thumbnail: bool = False,
    thumbnail_threshold: int = 1_000_000,
):
    """Generate parquet metadata, using the already-partitioned parquet files
    for this catalog.

    For more information on the general parquet metadata files, and why we write them, see
    https://arrow.apache.org/docs/python/parquet.html#writing-metadata-and-common-metadata-files

    Args:
        catalog_path (str): base path for the catalog
        order_by_healpix (bool): use False if the dataset is not to be reordered by
            breadth-first healpix pixel (e.g. secondary indexes)
        output_path (str): base path for writing out metadata files
            defaults to `catalog_path` if unspecified
        create_thumbnail (bool): if True, create a data thumbnail parquet file for
            the dataset. Defaults to True.
        thumbnail_threshold (int): maximum number of rows in the data thumbnail,
            which is otherwise one row/partition. Defaults to 1_000_000.

    Returns:
        sum of the number of rows in the dataset.
    """
    ignore_prefixes = [
        "intermediate",
        "_common_metadata",
        "_metadata",
        "data_thumbnail",
    ]

    catalog_path = get_upath(catalog_path)
    dataset_subdir = catalog_path / "dataset"
    (dataset_path, dataset) = file_io.read_parquet_dataset(
        dataset_subdir,
        ignore_prefixes=ignore_prefixes,
        exclude_invalid_files=True,
    )
    metadata_collector = []
    # Collect the healpix pixels so we can sort before writing.
    healpix_pixels = []
    total_rows = 0

    # Collect the first rows for the data thumbnail
    first_rows = []
    pq_file_list = set()
    if create_thumbnail:
        # The thumbnail_threshold threshold is the maximum number of Parquet rows
        # per pixel, used to prevent memory issues. It doesn't make sense for the
        # thumbnail to have more rows than this. If it does, randomly sample those
        # available.
        row_limit = min(len(dataset.files), thumbnail_threshold)
        # Create set for O(1) lookups in the loop that follows
        pq_file_list = set(random.sample(dataset.files, row_limit))

    for single_file in dataset.files:
        relative_path = single_file[len(dataset_path) + 1 :]
        file = file_io.read_parquet_file(dataset_subdir / relative_path)
        single_metadata = file.metadata

        # Users must set the file path of each chunk before combining the metadata.
        single_metadata.set_file_path(relative_path)

        if order_by_healpix:
            healpix_pixel = paths.get_healpix_from_path(relative_path)
            healpix_pixels.append(healpix_pixel)

        metadata_collector.append(single_metadata)
        total_rows += single_metadata.num_rows

        if create_thumbnail and single_file in pq_file_list:
            first_rows.append(next(file.iter_batches(batch_size=1)))

    ## Write out the two metadata files
    if output_path is None:
        output_path = catalog_path
    if order_by_healpix:
        argsort = get_pixel_argsort(healpix_pixels)
        metadata_collector = np.array(metadata_collector)[argsort]
    catalog_base_dir = get_upath(output_path)
    file_io.make_directory(catalog_base_dir / "dataset", exist_ok=True)
    metadata_file_pointer = paths.get_parquet_metadata_pointer(catalog_base_dir)
    common_metadata_file_pointer = paths.get_common_metadata_pointer(catalog_base_dir)

    file_io.write_parquet_metadata(
        dataset.schema,
        metadata_file_pointer,
        metadata_collector=metadata_collector,
        write_statistics=True,
    )
    file_io.write_parquet_metadata(dataset.schema, common_metadata_file_pointer)

    ## Write out the thumbnail file
    if create_thumbnail:
        data_thumbnail_pointer = paths.get_data_thumbnail_pointer(catalog_path)
        data_thumbnail = pa.Table.from_batches(first_rows, dataset.schema)
        if SPATIAL_INDEX_COLUMN in data_thumbnail.column_names:
            data_thumbnail = data_thumbnail.sort_by(SPATIAL_INDEX_COLUMN)
        with data_thumbnail_pointer.open("wb") as f_out:
            pq.write_table(data_thumbnail, f_out)

    return total_rows


def read_row_group_fragments(metadata_file: str):
    """Generator for metadata fragment row groups in a parquet metadata file.

    Args:
        metadata_file (str): path to `_metadata` file.
    """
    metadata_file = get_upath(metadata_file)
    if not file_io.is_regular_file(metadata_file):
        metadata_file = paths.get_parquet_metadata_pointer(metadata_file)

    dataset = pds.parquet_dataset(metadata_file, filesystem=metadata_file.fs)

    for frag in dataset.get_fragments():
        yield from frag.row_groups


def _nonemin(value1, value2):
    """Similar to numpy's nanmin, but excludes `None` values.

    NB: If both values are `None`, this will still return `None`"""
    if value1 is None:
        return value2
    if value2 is None:
        return value1
    return min(value1, value2)


def _nonemax(value1, value2):
    """Similar to numpy's nanmax, but excludes `None` values.

    NB: If both values are `None`, this will still return `None`"""
    if value1 is None:
        return value2
    if value2 is None:
        return value1
    return max(value1, value2)


def _pick_columns(
    first_row_group,
    exclude_hats_columns: bool = True,
    exclude_columns: list[str] = None,
    include_columns: list[str] = None,
    only_numeric_columns: bool = False,
):
    """Convenience method to find the desired columns and their indexes, given
    some conventional user preferences."""

    if include_columns is None:
        include_columns = []

    if exclude_columns is None:
        exclude_columns = []
    if exclude_hats_columns:
        exclude_columns.extend(["Norder", "Dir", "Npix", "_healpix_29"])
    num_columns = first_row_group.num_columns

    column_names = [
        first_row_group.column(col).path_in_schema for col in range(0, first_row_group.num_columns)
    ]
    numeric_columns = [
        first_row_group.column(col).path_in_schema
        for col in range(0, num_columns)
        if first_row_group.column(col).physical_type in ("DOUBLE", "FLOAT", "DECIMAL")
        or "INT" in first_row_group.column(col).physical_type
    ]

    column_names = [name.removesuffix(".list.element") for name in column_names]
    numeric_columns = [name.removesuffix(".list.element") for name in numeric_columns]
    good_column_indexes = [
        index
        for index, name in enumerate(column_names)
        if (len(include_columns) == 0 or name in include_columns)
        and not (len(exclude_columns) > 0 and name in exclude_columns)
        and (not only_numeric_columns or name in numeric_columns)
    ]
    column_names = [column_names[i] for i in good_column_indexes]

    return good_column_indexes, column_names


def aggregate_column_statistics(
    metadata_file: str | Path | UPath,
    exclude_hats_columns: bool = True,
    exclude_columns: list[str] = None,
    include_columns: list[str] = None,
    only_numeric_columns: bool = False,
    include_pixels: list[HealpixPixel] = None,
):
    """Read footer statistics in parquet metadata, and report on global min/max values.

    Args:
        metadata_file (str | Path | UPath): path to `_metadata` file
        exclude_hats_columns (bool): exclude HATS spatial and partitioning fields
            from the statistics. Defaults to True.
        exclude_columns (List[str]): additional columns to exclude from the statistics.
        include_columns (List[str]): if specified, only return statistics for the column
            names provided. Defaults to None, and returns all non-hats columns.
        include_pixels (list[HealpixPixel]): if specified, only return statistics
            for the pixels indicated. Defaults to none, and returns all pixels.
        only_numeric_columns (bool): only include columns that are numeric (integer or
            floating point) in the statistics. If True, the entire frame should be numeric.
    Returns:
        dataframe with global summary statistics
    """
    total_metadata = file_io.read_parquet_metadata(metadata_file)
    num_row_groups = total_metadata.num_row_groups
    first_row_group = total_metadata.row_group(0)

    good_column_indexes, column_names = _pick_columns(
        first_row_group=first_row_group,
        exclude_hats_columns=exclude_hats_columns,
        exclude_columns=exclude_columns,
        include_columns=include_columns,
        only_numeric_columns=only_numeric_columns,
    )
    if not good_column_indexes:
        return pd.DataFrame()

    extrema = None

    for row_group_index in range(0, num_row_groups):
        row_group = total_metadata.row_group(row_group_index)
        if include_pixels is not None:
            pixel = paths.get_healpix_from_path(row_group.column(0).file_path)
            if pixel not in include_pixels:
                continue
        row_stats = [
            (
                (None, None, 0, 0)
                if row_group.column(col).statistics is None
                else (
                    row_group.column(col).statistics.min,
                    row_group.column(col).statistics.max,
                    row_group.column(col).statistics.null_count,
                    row_group.column(col).num_values,
                )
            )
            for col in good_column_indexes
        ]
        if extrema is None:
            extrema = row_stats
        ## This is annoying, but avoids extra copies, or none comparison.
        else:
            extrema = [
                (
                    (_nonemin(extrema[col][0], row_stats[col][0])),
                    (_nonemax(extrema[col][1], row_stats[col][1])),
                    extrema[col][2] + row_stats[col][2],
                    extrema[col][3] + row_stats[col][3],
                )
                for col in range(0, len(good_column_indexes))
            ]

    if extrema is None:
        return pd.DataFrame()

    stats_lists = np.array(extrema).T

    frame = (
        pd.DataFrame(
            {
                "column_names": column_names,
                "min_value": stats_lists[0],
                "max_value": stats_lists[1],
                "null_count": stats_lists[2],
                "row_count": stats_lists[3],
            }
        )
        .set_index("column_names")
        .astype({"null_count": int, "row_count": int})
    )
    return frame


# pylint: disable=too-many-positional-arguments
def per_pixel_statistics(
    metadata_file: str | Path | UPath,
    exclude_hats_columns: bool = True,
    exclude_columns: list[str] = None,
    include_columns: list[str] = None,
    only_numeric_columns: bool = False,
    include_stats: list[str] = None,
    multi_index: bool = False,
    include_pixels: list[HealpixPixel] = None,
    per_row_group: bool = False,
):
    """Read footer statistics in parquet metadata, and report on statistics about
    each pixel partition.

    Args:
        metadata_file (str | Path | UPath): path to `_metadata` file
        exclude_hats_columns (bool): exclude HATS spatial and partitioning fields
            from the statistics. Defaults to True.
        exclude_columns (List[str]): additional columns to exclude from the statistics.
        include_columns (List[str]): if specified, only return statistics for the column
            names provided. Defaults to None, and returns all non-hats columns.
        only_numeric_columns (bool): only include columns that are numeric (integer or
            floating point) in the statistics. If True, the entire frame should be numeric.
        include_pixels (list[HealpixPixel]): if specified, only return statistics
            for the pixels indicated. Defaults to none, and returns all pixels.
        include_stats (List[str]): if specified, only return the kinds of values from list
            (min_value, max_value, null_count, row_count). Defaults to None, and returns all values.
        multi_index (bool): should the returned frame be created with a multi-index, first on
            pixel, then on column name? Default is False, and instead indexes on pixel, with
            separate columns per-data-column and stat value combination.
        per_row_group (bool): should the returned data be even more fine-grained and provide
            per row group (within each pixel) level statistics? Default is currently False,
            but is very likely to change.
    Returns:
        dataframe with granular per-pixel statistics
    """
    total_metadata = file_io.read_parquet_metadata(metadata_file)
    num_row_groups = total_metadata.num_row_groups
    first_row_group = total_metadata.row_group(0)

    good_column_indexes, column_names = _pick_columns(
        first_row_group=first_row_group,
        exclude_hats_columns=exclude_hats_columns,
        exclude_columns=exclude_columns,
        include_columns=include_columns,
        only_numeric_columns=only_numeric_columns,
    )
    if not good_column_indexes:
        return pd.DataFrame()

    all_stats = ["min_value", "max_value", "null_count", "row_count"]
    int_stats = ["null_count", "row_count"]

    if include_stats is None or len(include_stats) == 0:
        include_stats = all_stats
    else:
        for stat in include_stats:
            if stat not in all_stats:
                raise ValueError(f"include_stats must be from list {all_stats} (found {stat})")
        int_stats = [stat for stat in int_stats if stat in include_stats]

    stat_mask = np.array([ind for ind, stat in enumerate(all_stats) if stat in include_stats])
    combined_stats = {}
    pixels = []
    leaf_stats = []

    for row_group_index in range(0, num_row_groups):
        row_group = total_metadata.row_group(row_group_index)
        pixel = paths.get_healpix_from_path(row_group.column(0).file_path)
        if include_pixels is not None and pixel not in include_pixels:
            continue
        row_stats = [
            (
                [None, None, 0, 0]
                if row_group.column(col).statistics is None
                else [
                    row_group.column(col).statistics.min,
                    row_group.column(col).statistics.max,
                    row_group.column(col).statistics.null_count,
                    row_group.column(col).num_values,
                ]
            )
            for col in good_column_indexes
        ]
        if per_row_group:
            row_stats = np.take(row_stats, stat_mask, axis=1)
            pixels.append(pixel)
            leaf_stats.append(row_stats)
        else:
            if pixel not in combined_stats:
                combined_stats[pixel] = row_stats
            else:
                current_stats = combined_stats[pixel]
                combined_stats[pixel] = [
                    (
                        _nonemin(current_stats[i][0], row_stats[i][0]),
                        _nonemax(current_stats[i][1], row_stats[i][1]),
                        current_stats[i][2] + row_stats[i][2],
                        current_stats[i][3] + row_stats[i][3],
                    )
                    for i in range(0, len(good_column_indexes))
                ]

    if per_row_group:
        stats_lists = np.array(leaf_stats)
    else:
        pixels = list(combined_stats.keys())
        stats_lists = np.array(
            [np.take(row_stats, stat_mask, axis=1) for row_stats in combined_stats.values()]
        )
    original_shape = stats_lists.shape

    if len(stats_lists) == 0:
        return pd.DataFrame()

    if multi_index:
        stats_lists = stats_lists.reshape((original_shape[0] * original_shape[1], original_shape[2]))
        frame = pd.DataFrame(
            stats_lists,
            index=pd.MultiIndex.from_product([pixels, column_names], names=["pixel", "column"]),
            columns=include_stats,
        ).astype({stat_name: int for stat_name in int_stats})
    else:
        stats_lists = stats_lists.reshape((original_shape[0], original_shape[1] * original_shape[2]))
        mod_col_names = [[f"{col_name}: {stat}" for stat in include_stats] for col_name in column_names]
        mod_col_names = np.array(mod_col_names).flatten()
        int_col_names = [[f"{col_name}: {stat}" for stat in int_stats] for col_name in column_names]
        int_col_names = np.array(int_col_names).flatten()
        frame = pd.DataFrame(stats_lists, index=pixels, columns=mod_col_names).astype(
            {stat_name: int for stat_name in int_col_names}
        )
    return frame



from pathlib import Path
from typing import Iterable

import numpy as np
from upath import UPath

import hats.pixel_math.healpix_shim as hp
from hats.io import file_io, paths


def read_skymap(catalog, order):
    """Read the object spatial distribution information from a healpix skymap FITS file.

    Args:
        catalog: Catalog object corresponding to an on-disk catalog.
        order: healpix order to read the skymap at. If None, the order of the default
            skymap will be used. We will try to load from alternative skymap orders,
            where appropriate.
    Returns:
        one-dimensional numpy array of long integers where the value at each index
        corresponds to the number of objects found at the healpix pixel.
    """
    if order is not None and catalog.catalog_info.skymap_alt_orders:
        available_orders = catalog.catalog_info.skymap_alt_orders
        best_order_idx = np.searchsorted(available_orders, order)
        if best_order_idx < len(available_orders):
            best_order = available_orders[best_order_idx]

            ## We have a file with the same order - just use it
            if best_order == order:
                return file_io.read_fits_image(
                    paths.get_skymap_file_pointer(catalog_base_dir=catalog.catalog_base_dir, order=order)
                )

            ## We have a file with a greater order - downsample it
            skymap = file_io.read_fits_image(
                paths.get_skymap_file_pointer(catalog_base_dir=catalog.catalog_base_dir, order=best_order)
            )
            return skymap.reshape(hp.order2npix(order), -1).sum(axis=1)

    if catalog.catalog_info.skymap_order:
        if order is None or order == catalog.catalog_info.skymap_order:
            return file_io.read_fits_image(
                paths.get_skymap_file_pointer(catalog_base_dir=catalog.catalog_base_dir)
            )
        if order > catalog.catalog_info.skymap_order:
            raise ValueError(
                f"order should be less than stored skymap order ({catalog.catalog_info.skymap_order})"
            )
        skymap = file_io.read_fits_image(
            paths.get_skymap_file_pointer(catalog_base_dir=catalog.catalog_base_dir)
        )
        return skymap.reshape(hp.order2npix(order), -1).sum(axis=1)

    ## Deprecated - prefer reading skymap.fits to reading point_map.fits
    map_file_pointer = paths.get_point_map_file_pointer(catalog.catalog_base_dir)
    point_map = file_io.read_fits_image(map_file_pointer)
    point_map_order = hp.npix2order(len(point_map))
    if order is None or order == point_map_order:
        return point_map
    if point_map_order < order:
        raise ValueError(f"order should be less than stored skymap order ({point_map_order})")
    return point_map.reshape(hp.order2npix(order), -1).sum(axis=1)


def write_skymap(histogram: np.ndarray, catalog_dir: str | Path | UPath, orders: list | int | None = None):
    """Write the object spatial distribution information to a healpix SKYMAP FITS file.

    Args:
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        catalog_dir (path-like): base directory of the catalog in which to write the skymap file(s)
        order (list): list of orders to write additional skymap files. if provided and not empty,
            we will write a `skymap.K.fits` for each integer K in the list. if empty or None,
            we will not write additional files.
    """
    catalog_dir = file_io.get_upath(catalog_dir)
    map_file_pointer = paths.get_skymap_file_pointer(catalog_dir)
    file_io.write_fits_image(histogram=histogram, map_file_pointer=map_file_pointer)
    if orders:
        original_order = hp.npix2order(len(histogram))
        if not isinstance(orders, Iterable):
            ## allow input of a single order.
            orders = [orders]
        for order in orders:
            if order > original_order:
                raise ValueError(
                    f"sub-sampling skymap order should be less than overall order ({original_order})"
                )
            sampled_histogram = histogram.reshape(hp.order2npix(order), -1).sum(axis=1)
            map_file_pointer = paths.get_skymap_file_pointer(catalog_dir, order=order)
            file_io.write_fits_image(histogram=sampled_histogram, map_file_pointer=map_file_pointer)



import shutil

import pytest

import hats
from hats.catalog.catalog_collection import CatalogCollection
from hats.catalog.dataset.collection_properties import CollectionProperties
from hats.io.file_io import get_upath_for_protocol
from hats.loaders import read_hats


def test_read_hats_collection(small_sky_collection_dir, small_sky_order1_catalog):
    collection = read_hats(small_sky_collection_dir)
    assert isinstance(collection, CatalogCollection)
    assert collection.collection_path == small_sky_collection_dir
    assert collection.main_catalog_dir == small_sky_collection_dir / "small_sky_order1"
    assert collection.all_margins == ["small_sky_order1_margin"]
    assert collection.default_margin_catalog_dir == small_sky_collection_dir / "small_sky_order1_margin"
    assert collection.all_indexes == {"id": "small_sky_order1_id_index"}
    assert collection.default_index_field == "id"
    assert collection.default_index_catalog_dir == small_sky_collection_dir / "small_sky_order1_id_index"
    assert collection.get_healpix_pixels() == small_sky_order1_catalog.get_healpix_pixels()


def test_read_hats_collection_main_catalog_invalid(small_sky_collection_dir, tmp_path):
    """Test that the main catalog is of the correct `Catalog` type"""
    collection_base_dir = tmp_path / "collection"
    shutil.copytree(small_sky_collection_dir, collection_base_dir)
    assert collection_base_dir.exists()
    collection_properties = CollectionProperties.read_from_dir(collection_base_dir)
    collection_properties.hats_primary_table_url = "small_sky_order1_margin"
    collection_properties.to_properties_file(collection_base_dir)
    with pytest.raises(TypeError):
        read_hats(collection_base_dir)


def test_read_hats_default_margin_not_specified(small_sky_collection_dir, tmp_path):
    collection_base_dir = tmp_path / "collection"
    shutil.copytree(small_sky_collection_dir, collection_base_dir)
    assert collection_base_dir.exists()
    collection_properties = CollectionProperties.read_from_dir(collection_base_dir)
    collection_properties.default_margin = None
    collection_properties.to_properties_file(collection_base_dir)
    collection = read_hats(collection_base_dir)
    assert collection.default_margin_catalog_dir is None


def test_read_hats_default_index_not_specified(small_sky_collection_dir, tmp_path):
    collection_base_dir = tmp_path / "collection"
    shutil.copytree(small_sky_collection_dir, collection_base_dir)
    assert collection_base_dir.exists()
    collection_properties = CollectionProperties.read_from_dir(collection_base_dir)
    collection_properties.default_index = None
    collection_properties.to_properties_file(collection_base_dir)
    collection = read_hats(collection_base_dir)
    assert collection.default_index_catalog_dir is None


def test_read_hats_index_dir_for_field(small_sky_collection_dir, tmp_path):
    collection_base_dir = tmp_path / "collection"
    shutil.copytree(small_sky_collection_dir, collection_base_dir)
    assert collection_base_dir.exists()

    collection = read_hats(collection_base_dir)

    # If no field specified, the default index dir is returned
    assert collection.default_index_field == "id"
    assert collection.get_index_dir_for_field() == collection.get_index_dir_for_field("id")

    # There are indexes but none match the field name
    with pytest.raises(ValueError, match="not specified"):
        collection.get_index_dir_for_field("name")

    # There are no indexes available
    collection_properties = CollectionProperties.read_from_dir(collection_base_dir)
    collection_properties.all_indexes = None
    collection_properties.default_index = None
    collection_properties.to_properties_file(collection_base_dir)
    collection = read_hats(collection_base_dir)
    with pytest.raises(ValueError, match="not specified"):
        collection.get_index_dir_for_field("id")


def test_read_hats_collection_info_only(collection_path):
    with pytest.raises(FileNotFoundError):
        read_hats(collection_path)


def test_read_hats_branches(
    small_sky_dir,
    small_sky_order1_dir,
    association_catalog_path,
    small_sky_source_object_index_dir,
    margin_catalog_path,
    small_sky_source_dir,
    test_data_dir,
):
    read_hats(small_sky_dir)
    read_hats(small_sky_order1_dir)
    read_hats(association_catalog_path)
    read_hats(small_sky_source_object_index_dir)
    read_hats(margin_catalog_path)
    read_hats(small_sky_source_dir)
    read_hats(test_data_dir / "square_map")


def test_read_hats_initializes_upath_once(small_sky_dir, mocker):
    mock_method = "hats.io.file_io.file_pointer.get_upath_for_protocol"
    # Setting the side effect allows us to run the mocked function's code
    mocked_upath_call = mocker.patch(mock_method, side_effect=get_upath_for_protocol)
    read_hats(small_sky_dir)
    # The construction of the UPath is called once, at the start of `read_hats`
    mocked_upath_call.assert_called_once_with(small_sky_dir)


def test_read_hats_with_s3():
    upath = get_upath_for_protocol("s3://bucket/catalog")
    assert upath.storage_options.get("anon")
    assert upath.storage_options["default_block_size"] == 32 * 1024


def test_read_hats_with_http():
    upath_http = get_upath_for_protocol("http://catalog")
    assert upath_http.fs.block_size == 32 * 1024
    upath_https = get_upath_for_protocol("https://catalog")
    assert upath_https.fs.block_size == 32 * 1024


def test_read_hats_nonstandard_npix_suffix(
    small_sky_npix_alt_suffix_dir,
    small_sky_npix_as_dir_dir,
):
    read_hats(small_sky_npix_alt_suffix_dir)
    read_hats(small_sky_npix_as_dir_dir)


def test_read_hats_original_schema(small_sky_order1_dir):
    cat = hats.read_hats(small_sky_order1_dir)
    assert cat.schema == cat.original_schema



import numpy as np
import pytest

import hats.pixel_math.healpix_shim as hp
from hats import read_hats
from hats.io.file_io import read_fits_image
from hats.io.skymap import read_skymap, write_skymap


def test_write_skymap_roundtrip(tmp_path):
    """Test the reading/writing of a catalog point map"""
    dense = np.arange(0, 3072)

    # Writes a single, non-sampled, skymap
    write_skymap(dense, tmp_path)
    skymap_path = tmp_path / "skymap.fits"
    assert skymap_path.exists()
    skymap_order1_path = tmp_path / "skymap.1.fits"
    assert not skymap_order1_path.exists()
    counts_skymap = read_fits_image(tmp_path / "skymap.fits")
    np.testing.assert_array_equal(counts_skymap, dense)


@pytest.mark.parametrize("orders", [[], [2], [1, 2], [2, 1], [1], [0], [0, 1, 2, 3, 4]])
def test_write_sampled_skymaps_roundtrip(tmp_path, orders):
    """Test the reading/writing of a catalog point map"""
    dense = np.arange(0, 3072)
    assert sum(dense) == 4_717_056

    write_skymap(dense, tmp_path, orders)
    skymap_path = tmp_path / "skymap.fits"
    assert skymap_path.exists()
    counts_skymap = read_fits_image(tmp_path / "skymap.fits")
    np.testing.assert_array_equal(counts_skymap, dense)

    for order in range(0, 5):
        skymap_atorder_path = tmp_path / f"skymap.{order}.fits"
        if order in orders:
            assert skymap_atorder_path.exists()
            read_histogram = read_fits_image(skymap_atorder_path)
            assert hp.npix2order(len(read_histogram)) == order
            # Check that the contents have same overall count, and we're summing quadratically.
            assert sum(read_histogram) == 4_717_056
            pixel_size = 4 ** (4 - order)
            assert read_histogram[-1] == sum(dense[-1 * pixel_size :])
        else:
            assert not skymap_atorder_path.exists()


def test_write_sampled_skymaps_error(tmp_path):
    """Test the reading/writing of a catalog point map"""
    dense = np.arange(0, 3072)

    with pytest.raises(ValueError, match="order should be less"):
        write_skymap(dense, tmp_path, 13)


def test_read_alt_skymap(small_sky_source_dir, mocker):
    """Test that we're reading the file we expect to read, and get the
    appropriate length for the returned skymap histogram."""
    catalog = read_hats(small_sky_source_dir)

    mock_method = "hats.io.file_io.read_fits_image"
    # Setting the side effect allows us to run the mocked function's code
    mocked_read_fits_call = mocker.patch(mock_method, side_effect=read_fits_image)

    ## Requesting no order should use the default skymap (order 6)
    skymap = read_skymap(catalog, None)
    assert len(skymap) == 49152

    mocked_read_fits_call.assert_called_once()
    path_arg = mocked_read_fits_call.call_args.args[0]
    assert str(small_sky_source_dir / "skymap.fits") == str(path_arg)

    mocked_read_fits_call.reset_mock()

    ## Requesting available alt value should use that skymap
    skymap = read_skymap(catalog, 4)
    assert len(skymap) == 3_072

    mocked_read_fits_call.assert_called_once()
    path_arg = mocked_read_fits_call.call_args.args[0]
    assert str(small_sky_source_dir / "skymap.4.fits") == str(path_arg)

    mocked_read_fits_call.reset_mock()

    ## Requesting unavailable alt value should use next-highest alt skymap
    skymap = read_skymap(catalog, 3)
    assert len(skymap) == 768

    mocked_read_fits_call.assert_called_once()
    path_arg = mocked_read_fits_call.call_args.args[0]
    assert str(small_sky_source_dir / "skymap.4.fits") == str(path_arg)

    mocked_read_fits_call.reset_mock()

    ## Requesting too-high alt-value should use the default skymap
    skymap = read_skymap(catalog, 5)
    assert len(skymap) == 12_288

    mocked_read_fits_call.assert_called_once()
    path_arg = mocked_read_fits_call.call_args.args[0]
    assert str(small_sky_source_dir / "skymap.fits") == str(path_arg)

    mocked_read_fits_call.reset_mock()


def test_read_noalt_skymap(small_sky_source_dir, mocker):
    """Test that we're reading the file we expect to read, and get the
    appropriate length for the returned skymap histogram."""
    catalog = read_hats(small_sky_source_dir)
    ## Alternate sky map orders are only used when present in the properties file.
    catalog.catalog_info.skymap_alt_orders = None

    mock_method = "hats.io.file_io.read_fits_image"
    # Setting the side effect allows us to run the mocked function's code
    mocked_read_fits_call = mocker.patch(mock_method, side_effect=read_fits_image)

    ## Requesting no order should use the default skymap (order 6)
    skymap = read_skymap(catalog, None)
    assert len(skymap) == 49_152

    mocked_read_fits_call.assert_called_once()
    path_arg = mocked_read_fits_call.call_args.args[0]
    assert str(small_sky_source_dir / "skymap.fits") == str(path_arg)

    mocked_read_fits_call.reset_mock()

    ## Requesting (otherwise-)available alt value will still use default
    skymap = read_skymap(catalog, 4)
    assert len(skymap) == 3_072

    mocked_read_fits_call.assert_called_once()
    path_arg = mocked_read_fits_call.call_args.args[0]
    assert str(small_sky_source_dir / "skymap.fits") == str(path_arg)

    mocked_read_fits_call.reset_mock()

    ## Requesting waaay too-high order should error.
    with pytest.raises(ValueError, match="order should be less"):
        skymap = read_skymap(catalog, 13)
